#include <OpenHome/Stream.h>

#include <string.h>

using namespace OpenHome;

// Sxx

Sxx::Sxx(TUint aMaxBytes)
    : iMaxBytes(aMaxBytes)
    , iBytes(0)
{
}

Sxx::~Sxx()
{
}

// Srx

Srx::Srx(TUint aMaxBytes, IReaderSource& aSource)
    : Sxx(aMaxBytes)
    , iSource(aSource)
    , iOffset(0)
{
}

Brn Srx::Read(TUint aBytes)
{
    TByte* ptr = Ptr();

    if (iBytes - iOffset < aBytes) {  // read not satisfied from data already in the buffer
        if (aBytes > iMaxBytes) { // unable to store the requested number of bytes in the buffer
		    THROW(ReaderError);
        }

        if (iMaxBytes - iOffset < aBytes) { // unable to fit requested bytes after current offset
            (void)memmove(ptr, ptr + iOffset, iBytes - iOffset); // so make some more room
            iBytes -= iOffset;
            iOffset = 0;
        }

        while (iBytes - iOffset < aBytes) { // collect more data from the source
            Bwn buffer(ptr + iBytes, iMaxBytes - iBytes);
            iSource.Read(buffer);
            iBytes += buffer.Bytes();
        }
    }

    TUint offset = iOffset;
    iOffset += aBytes;
    return (Brn(ptr + offset, aBytes));
}

Brn Srx::ReadUntil(TByte aSeparator)
{
    TByte* ptr = Ptr();
    
    TByte* start = ptr + iOffset;
    TByte* current = start;
    
    TUint count = 0;
    TUint remaining = iBytes - iOffset;

    for (;;) {
        while (remaining) {
            if (*current++ == aSeparator) {
                iOffset += count + 1; // skip over the separator
                return (Brn(start, count));
            }
            count++;
            remaining--;
        }
    
        // separator not found in current buffer
        
        if (iOffset > 0) {   // so move everything down
            start -= iOffset;
            iBytes -= iOffset;
            current -= iOffset;
            if (iBytes) {
                memmove(ptr, ptr + iOffset, iBytes);
            }
            iOffset = 0;
        }
        
        if (iBytes >= iMaxBytes) { // buffer full and no separator
        	THROW(ReaderError);
        }
        
        Bwn buffer(ptr + iBytes, iMaxBytes - iBytes); // collect more data from the source

        iSource.Read(buffer);
        
        TUint additional = buffer.Bytes();
        
        iBytes += additional;
        remaining += additional;
    }
}

void Srx::ReadFlush()
{
    iBytes = 0;
    iOffset = 0;
    iSource.ReadFlush();
}

void Srx::ReadInterrupt()
{
    iSource.ReadInterrupt();
}

Brn Srx::Snaffle()
{
    Brn rem(Ptr() + iOffset, iBytes - iOffset);
    iOffset = 0;
    iBytes = 0;
    return rem;
}


// Swx

Swx::Swx(TUint aMaxBytes, IWriter& aWriter)
    : Sxx(aMaxBytes) 
    , iWriter(aWriter)
{
}

void Swx::Error()
{
    iBytes = 0;
    THROW(WriterError);
}

void Swx::Write(TByte aValue)
{
    TByte* ptr = Ptr();
    
    if (iBytes >= iMaxBytes) { // would overflow, flush the buffer
        iWriter.Write(Brn(ptr, iBytes));
        iBytes = 0;
    }

    *(ptr + iBytes++) = aValue;
}

void Swx::Write(const Brx& aBuffer)
{
    TByte* ptr = Ptr();
    
    TUint bytes = aBuffer.Bytes();
    
    if (iBytes + bytes > iMaxBytes) { // would overflow, drain the buffer
    
        WriteDrain();
        
        if (bytes > iMaxBytes) { // would still overflow
            iWriter.Write(aBuffer); // pass it on
        }
    }
    
    memcpy(ptr + iBytes, aBuffer.Ptr(), bytes);
    
    iBytes += bytes;
}

void Swx::WriteDrain()
{
    if (iBytes) {
        TByte* ptr = Ptr();
        iWriter.Write(Brn(ptr, iBytes));
        iBytes = 0;
    }
}

void Swx::WriteFlush()
{
    WriteDrain();
    iWriter.WriteFlush();
}

// Swp (Parasite on a host read stream)

Swp::Swp(Srx& aHost, IWriter& aWriter)
    : Swx(aHost.iMaxBytes, aWriter)
    , iHost(aHost)
{
}

TByte* Swp::Ptr()
{
    return (iHost.Ptr());
}

Swp::~Swp()
{
}

// ReaderBuffer

ReaderBuffer::ReaderBuffer(const Brx& aBuffer)
{
	Set(aBuffer);
}

void ReaderBuffer::Set(const Brx& aBuffer)
{
    iBuffer.Set(aBuffer);
    ReadFlush();
}

Brn ReaderBuffer::Read(TUint aBytes)
{
    TUint offset = iOffset + aBytes;
    
    if (offset > iBuffer.Bytes()) {
        THROW(ReaderError);
    }
    
    Brn result = iBuffer.Split(iOffset, aBytes);
    
    iOffset = offset;
    
    return (result);
}

TUint ReaderBuffer::Bytes() const
{   
    return (iBuffer.Bytes() - iOffset);
}

Brn ReaderBuffer::ReadRemaining()
{
    Brn result = iBuffer.Split(iOffset);

    iOffset = iBuffer.Bytes();
    
    return (result);
}

Brn ReaderBuffer::ReadPartial(TUint aBytes)
{
    TUint remaining = iBuffer.Bytes() - iOffset;
    
    if (remaining == 0) {
        THROW(ReaderError);
    }
    
    TUint bytes = aBytes;
    
    if (bytes > remaining) {
        bytes = remaining;
    }

    Brn result = iBuffer.Split(iOffset, bytes);
    
    iOffset += bytes;
    
    return (result);
}

Brn ReaderBuffer::ReadUntil(TByte aSeparator)
{
    TUint start = iOffset;
    TUint offset = iOffset;
    
    TUint bytes = iBuffer.Bytes();
    
    while (offset < bytes) {
        if (iBuffer[offset++] == aSeparator) {
            iOffset = offset;
            return (iBuffer.Split(start, offset - start - 1));
        }
    }
    
    THROW(ReaderError);

	return (Brn(0,0));
}

void ReaderBuffer::ReadFlush()
{
    iOffset = 0;
}

void ReaderBuffer::ReadInterrupt()
{
}



// WriterBuffer

WriterBuffer::WriterBuffer(Bwx& aBuffer) : iBuffer(aBuffer)
{
}

void WriterBuffer::Write(TByte aValue)
{
    iBuffer.Append(aValue);
}

void WriterBuffer::Write(const Brx& aBuffer)
{
    iBuffer.Append(aBuffer);
}

void WriterBuffer::WriteFlush()
{
}

void WriterBuffer::Flush()
{
    iBuffer.Clear();
}


// ReaderBinary

ReaderBinary::ReaderBinary(IReader& aReader)
    : iReader(aReader)
{
}

const Brn ReaderBinary::Read(TUint aBytes)
{
    return (iReader.Read(aBytes));    
}

TUint ReaderBinary::ReadUintBe(TUint aBytes)
{
    OHASSERT(aBytes > 0);
    OHASSERT(aBytes <= sizeof(TUint));
    TUint val = 0;
    TUint count = 0;
    while(count < aBytes) {
        TUint byte = iReader.Read(1).At(0);
        val += byte << ((aBytes - count - 1)*8);
        count++;
    }
    return val;
}

TInt ReaderBinary::ReadIntBe(TUint aBytes)
{
    OHASSERT(aBytes > 0);
    OHASSERT(aBytes <= sizeof(TInt));
    TInt val = 0;
    TUint count = 0;

    TInt sbyte = iReader.Read(1).At(0);
    val += sbyte << ((aBytes - count - 1) * 8);
    count++;

    while(count < aBytes) {
        TUint byte = iReader.Read(1).At(0);
        val += byte << ((aBytes - count - 1) * 8);
        count++;
    }
    return val;
}

TUint ReaderBinary::ReadUintLe(TUint aBytes)
{
    OHASSERT(aBytes > 0);
    OHASSERT(aBytes <= sizeof(TUint));
    TUint val = 0;
    TUint shift = 0;
    while(shift < aBytes) {
        TUint byte = iReader.Read(1).At(0);
        val += byte << (shift*8);
        shift++;
    }
    return val;
}

TUint64 ReaderBinary::ReadUint64Be(TUint aBytes)
{
    OHASSERT(aBytes > 0);
    OHASSERT(aBytes <= sizeof(TUint64));
    TUint64 val = 0;
    TUint count = 0;
    while(count < aBytes) {
        TUint byte = iReader.Read(1).At(0);
        val += byte << ((aBytes - count - 1)*8);
        count++;
    }
    return val;
}

TUint64 ReaderBinary::ReadUint64Le(TUint aBytes)
{
    OHASSERT(aBytes > 0);
    OHASSERT(aBytes <= sizeof(TUint64));
    TUint64 val = 0;
    TUint shift = 0;
    while(shift < aBytes) {
        TUint byte = iReader.Read(1).At(0);
        val += byte << (shift*8);
        shift++;
    }
    return val;
}

// WriterBinary

WriterBinary::WriterBinary(IWriter& aWriter)
    : iWriter(aWriter)
{
}

void WriterBinary::Write(const Brx& aBuf)
{
    iWriter.Write(aBuf);
}

void WriterBinary::WriteUint8(TUint aValue)
{
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteUint16Be(TUint aValue)
{
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteUint24Be(TUint aValue)
{
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteUint32Be(TUint aValue)
{
    iWriter.Write((TByte)(aValue >> 24));
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteUint64Be(TUint64 aValue)
{
    iWriter.Write((TByte)(aValue >> 56));
    iWriter.Write((TByte)(aValue >> 48));
    iWriter.Write((TByte)(aValue >> 40));
    iWriter.Write((TByte)(aValue >> 32));
    iWriter.Write((TByte)(aValue >> 24));
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteInt8(TInt aValue)
{
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteInt16Be(TInt aValue)
{
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteInt24Be(TInt aValue)
{
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteInt32Be(TInt aValue)
{
    iWriter.Write((TByte)(aValue >> 24));
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}

void WriterBinary::WriteInt64Be(TInt64 aValue)
{
    iWriter.Write((TByte)(aValue >> 56));
    iWriter.Write((TByte)(aValue >> 48));
    iWriter.Write((TByte)(aValue >> 40));
    iWriter.Write((TByte)(aValue >> 32));
    iWriter.Write((TByte)(aValue >> 24));
    iWriter.Write((TByte)(aValue >> 16));
    iWriter.Write((TByte)(aValue >> 8));
    iWriter.Write((TByte)(aValue));
}


